/* This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/** global definitions and methods (code)
 *  @file
 *  @author King Kévin <kingkevin@cuvoodoo.info>
 *  @date 2016-2020
 */
/* standard libraries */
#include <stdint.h> // standard integer types
#include <stdlib.h> // general utilities
#include <string.h> // memory utilities

/* STM32 (including CM3) libraries */
#include <libopencmsis/core_cm3.h> // Cortex M3 utilities
#include <libopencm3/cm3/nvic.h> // interrupt handler
#include <libopencm3/cm3/systick.h> // SysTick library
#include <libopencm3/stm32/rcc.h> // real-time control clock library
#include <libopencm3/stm32/gpio.h> // general purpose input output library
#include <libopencm3/stm32/exti.h> // external interrupt defines

#include "global.h" // common methods

volatile bool user_input_available = false;

static volatile uint8_t user_input_buffer[64] = {0}; /**< ring buffer for received data */
static volatile uint8_t user_input_i = 0; /**< current position of read received data */
static volatile uint8_t user_input_used = 0; /**< how much data has been received and not red */

static volatile uint32_t sleep_duration = 0; /**< sleep duration count down (in SysTick interrupts) */

char* b2s(uint64_t binary, uint8_t rjust)
{
	static char string[64 + 1] = {0}; // the string representation to return
	uint8_t bit = LENGTH(string) - 1; // the index of the bit to print
	string[bit--] = '\0'; // terminate string

	while (binary) {
		if (binary & 1) {
			string[bit--] = '1';
		} else {
			string[bit--] = '0';
		}
		binary >>= 1;
	}

	while (64 - bit - 1 < rjust && bit > 0) {
		string[bit--] = '0';
	}

	return string;
}

/** switch on board LED */
inline void led_on(void)
{
#if defined(LED_PIN)
#if defined(LED_ON) && LED_ON
	gpio_set(GPIO_PORT(LED_PIN), GPIO_PIN(LED_PIN));
#else
	gpio_clear(GPIO_PORT(LED_PIN), GPIO_PIN(LED_PIN));
#endif // LED_ON
#endif // LED_PIN
}

/** switch off board LED */
inline void led_off(void)
{
#if defined(LED_PIN)
#if defined(LED_ON) && LED_ON
	gpio_clear(GPIO_PORT(LED_PIN), GPIO_PIN(LED_PIN));
#else
	gpio_set(GPIO_PORT(LED_PIN), GPIO_PIN(LED_PIN));
#endif // LED_ON
#endif // LED_PIN
}

/** toggle board LED */
inline void led_toggle(void)
{
#if defined(LED_PIN)
	gpio_toggle(GPIO_PORT(LED_PIN), GPIO_PIN(LED_PIN));
#endif // LED_PIN
}

void sleep_us(uint32_t duration)
{
	if (duration <= 5) { // less than the setup time
		for (volatile uint32_t nop = 0; nop < 5 * duration; nop++); // busy loop, approximate, hand tuned for 72 MHz system clock
		return;
	}

	duration -= 5; // subtract setup time
	systick_counter_disable(); // disable SysTick to reconfigure it
	if (!systick_set_frequency(1000000, rcc_ahb_frequency)) { // set SysTick frequency to microseconds
		 while (true); // unhandled error
	}
	systick_clear(); // reset SysTick
	systick_interrupt_enable(); // enable interrupt to count duration
	sleep_duration = duration; // save sleep duration for count down
	systick_counter_enable(); // start counting
	while (sleep_duration > 0) { // wait for count down to complete
		__WFI(); // go to sleep
	}
}

void sleep_ms(uint32_t duration)
{
	if (0 == duration) {
		return;
	}

	systick_counter_disable(); // disable SysTick to reconfigure it
	if (!systick_set_frequency(1000, rcc_ahb_frequency)) { // set SysTick frequency to milliseconds
		while (true); // unhandled error
	}
	systick_clear(); // reset SysTick
	systick_interrupt_enable(); // enable interrupt to count duration
	sleep_duration = duration; // save sleep duration for count down
	systick_counter_enable(); // start counting
	while (sleep_duration > 0) { // wait for count down to complete
		__WFI(); // go to sleep
	}
}

/** SysTick interrupt handler */
void sys_tick_handler(void)
{
	if (sleep_duration > 0) {
		sleep_duration--; // decrement duration
	}
	if (0 == sleep_duration) { // sleep complete
		systick_counter_disable(); // stop systick
		systick_interrupt_disable(); // stop interrupting
		sleep_duration = 0; // ensure it still is at 0
	}
}

char user_input_get(void)
{
	while (0 == user_input_used) { // wait for user input to be available (don't trust user_input_available since it user modifiable)
		__WFI(); // go to sleep
	}
	volatile char to_return = user_input_buffer[user_input_i]; // get the next available character
	user_input_i = (user_input_i + 1) % LENGTH(user_input_buffer); // update used buffer
	user_input_used--; // update used buffer
	user_input_available = (user_input_used != 0); // update available data
	return to_return;
}

void user_input_store(char c)
{
	// only save data if there is space in the buffer
	if (user_input_used >= LENGTH(user_input_buffer)) { // if buffer is full
		user_input_i = (user_input_i + 1) % LENGTH(user_input_buffer); // drop oldest data
		user_input_used--; // update used buffer information
	}
	user_input_buffer[(user_input_i + user_input_used) % LENGTH(user_input_buffer)] = c; // put character in buffer
	user_input_used++; // update used buffer
	user_input_available = true; // update available data
}

void board_setup(void)
{
#if defined(LED_PIN)
	// setup LED
	rcc_periph_clock_enable(GPIO_RCC(LED_PIN)); // enable clock for LED
	gpio_set_mode(GPIO_PORT(LED_PIN), GPIO_MODE_OUTPUT_2_MHZ, GPIO_CNF_OUTPUT_OPENDRAIN, GPIO_PIN(LED_PIN)); // LED it controller by pMOS by default pulled up
	led_off(); // switch off LED per default
#endif // LED_PIN

	// reset user input buffer
	user_input_available = false;
	user_input_i = 0;
	user_input_used = 0;
}
