/* This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/** library to communicate using USB-C controller FUSB302
 *  @file
 *  @author King Kévin <kingkevin@cuvoodoo.info>
 *  @date 2020
 *  @note peripherals used: I²C @ref usb_fusb302_i2c
 */
/* standard libraries */
#include <stdint.h> // standard integer types
#include <stdlib.h> // general utilities
#include <string.h> // memory utilities

/* STM32 (including CM3) libraries */
#include <libopencm3/stm32/rcc.h> // real-time control clock library
#include <libopencm3/stm32/gpio.h> // general purpose input output library
#include <libopencm3/stm32/i2c.h> // I²C library
#include <libopencm3/stm32/crc.h> // CRC utilities

/* own libraries */
#include "global.h" // global utilities
#include "usb_fusb302.h" // own devifition
#include "i2c_master.h" // I²C header and definitions

/* external library */
#include "usb_pd.h" // USB Power Delivery definitions

/** I²C peripheral use for communication with FUSB302
 *  @def usb_fusb302_i2c
 */
#define USB_FUSB302_I2C I2C1
/** FUSB302 I²C slave address 
 *  possible I²C slave addresses (provided by technical support):
 *  0x22: FUSB302BUCX, FUSB302BMPX, FUSB302BVMPX
 *  0x23: FUSB302B01MPX
 *  0x24: FUSB302B10MPX
 *  0x25: FUSB302B11MPX
 */
#define USB_FUSB302_ADDR 0x22

int16_t usb_fusb302_setup(void)
{
	rcc_periph_clock_enable(RCC_CRC); // enable cock for CRC domain to calculate in hardware the CRC
	
	i2c_master_setup(USB_FUSB302_I2C, 100); // setup I²C peripheral, with 100 kHz clock for compatibility with other devices (can go up to 1 MHz, but other slaves on the bus might not be able to go this fast)

	// test if FUSB302 is on the I²C bus
	enum i2c_master_rc i2c_rc = i2c_master_start(USB_FUSB302_I2C);
	if (I2C_MASTER_RC_NONE == i2c_rc) {
		i2c_rc = i2c_master_select_slave(USB_FUSB302_I2C, USB_FUSB302_ADDR, false, false);
		if (I2C_MASTER_RC_NONE != i2c_rc) {
			i2c_master_stop(USB_FUSB302_I2C);
			return -i2c_rc;
		}
	} else {
		i2c_master_stop(USB_FUSB302_I2C);
		return -100 - i2c_rc;
	}
	i2c_rc = i2c_master_stop(USB_FUSB302_I2C);
	if (I2C_MASTER_RC_NONE != i2c_rc) {
		return -200 - i2c_rc;
	}

	// do a software reset to have the register return to the default values
	int16_t rc = usb_fusb302_register_write(FUSB302_REG_RESET, FUSB302_REG_RESET_SW_RES);
	if (rc < 0) {
		return rc - 300;
	}

	// enable all power
	// not very energy efficient, but at least we don't have surprises
	const uint8_t reg = FUSB302_REG_POWER_OSCILLATOR | FUSB302_REG_POWER_MEASURE | FUSB302_REG_POWER_RECEIVER_REFERENCES | FUSB302_REG_POWER_BANDGAP_WAKEUP;
	rc = usb_fusb302_register_write(FUSB302_REG_POWER, reg);
	if (rc < 0) {
		return rc - 400;
	}

	return usb_fusb302_disconnect();
}

int16_t usb_fusb302_disconnect(void)
{
	int16_t reg;

	// disconnect all from CC pins
	reg = usb_fusb302_register_write(FUSB302_REG_SWITCHES0, 0);
	if (reg < 0) {
		return reg - 1000;
	}

	// disconnect VBUS
	reg = usb_fusb302_register_read(FUSB302_REG_MEASURE);
	if (reg < 0) {
		return reg - 1100;
	}
	reg &= ~FUSB302_REG_MEASURE_MEAS_VBUS;
	reg = usb_fusb302_register_write(FUSB302_REG_MEASURE, reg);
	if (reg < 0) {
		return reg - 1200;
	}

	return 0;
}

int16_t usb_fusb302_register_read(uint8_t address)
{
	uint8_t reg;
	enum i2c_master_rc rc = i2c_master_address_read(USB_FUSB302_I2C, USB_FUSB302_ADDR, false, &address, 1, &reg, 1);
	if (I2C_MASTER_RC_NONE == rc) {
		return reg;
	} else {
		return -rc;
	}
}

int16_t usb_fusb302_register_write(uint8_t address, uint8_t value)
{
	enum i2c_master_rc rc = i2c_master_address_write(USB_FUSB302_I2C, USB_FUSB302_ADDR, false, &address, 1, &value, 1);
	if (I2C_MASTER_RC_NONE == rc) {
		return 0;
	} else {
		return -rc;
	}
}

char usb_fusb302_version(void)
{
	int16_t reg = usb_fusb302_register_read(FUSB302_REG_DEVICE_ID); // read version register
	if (reg < 0) {
		return '!';
	}
	reg >>= 4; // just get the version
	if (reg < 0x08) { // undefined value
		return '?';
	} else {
		return reg - 0x08 + 'A';
	}
}

char usb_fusb302_revision(void)
{
	int16_t reg = usb_fusb302_register_read(FUSB302_REG_DEVICE_ID); // read version register
	if (reg < 0) {
		return -reg;
	} else {
		return (reg & 0x0f) + 'A';
	}
}

int32_t usb_fusb302_voltage(uint8_t source)
{
	if (source > 2) {
		return -1;
	}
	int16_t reg; // to store register values

	// this is not really explained in the datasheet, but power needs to be enabled first
	reg = FUSB302_REG_POWER_MEASURE | FUSB302_REG_POWER_RECEIVER_REFERENCES | FUSB302_REG_POWER_BANDGAP_WAKEUP;
	reg = usb_fusb302_register_write(FUSB302_REG_POWER, reg);
	if (reg < 0) {
		return reg - 100;
	}

	// set CC source
	reg = 0; // start with all disconnected
	if (1 == source) {
		reg |= FUSB302_REG_SWITCHES0_MEAS_CC1;
	} else if (2 == source) {
		reg |= FUSB302_REG_SWITCHES0_MEAS_CC2;
	}
	reg = usb_fusb302_register_write(FUSB302_REG_SWITCHES0, reg);
	if (reg < 0) {
		return reg - 200;
	}

	// increment MDAC until COMP is set
	uint8_t mdac; // DAC threshold
	for (mdac = 0; mdac < 0x3f; mdac++) {
		reg = mdac;
		if (0 == source) { // set if VBUS is the source, or CC
			reg |= FUSB302_REG_MEASURE_MEAS_VBUS;
		}
		reg = usb_fusb302_register_write(FUSB302_REG_MEASURE, reg);
		if (reg < 0) {
			return reg - 400;
		}
		sleep_us(100); // before that the result is wrong
		reg = usb_fusb302_register_read(FUSB302_REG_STATUS0);
		if (reg < 0) {
			return reg - 500;
		}
		if (0 == (reg & FUSB302_REG_STATUS0_COMP)) { // we passed the threshold
			break;
		}
	}

	// disconnected CC and VBUS
	reg = usb_fusb302_disconnect();
	if (reg < 0) {
		return reg;
	}

	if (0 == source) {
		return mdac * 420;
	} else {
		return mdac * 42;
	}
}

int16_t usb_fusb302_r(uint8_t cc)
{
	if (0 == cc || cc > 2) {
		return -2;
	}
	int16_t reg; // to store register values
	int16_t r = -1; // the resistor value to return

	// set PU current (HOST_CUR) to 330 uA to be able to differentiate Ra and Rd at once using BC_LEVEL and COMP
	reg = usb_fusb302_register_read(FUSB302_REG_STATUS0);
	if (reg < 0) {
		return reg - 100;
	}
	reg &= ~FUSB302_REG_CONTROL0_HOST_CUR_MASK;
	reg |= FUSB302_REG_CONTROL0_HOST_CUR_330UA;
	reg = usb_fusb302_register_write(FUSB302_REG_CONTROL0, reg);
	if (reg < 0) {
		return reg - 200;
	}

	// set MDAC to 2.6V (330 uA * 5.1k * 120% = 2.02V)
	reg = usb_fusb302_register_write(FUSB302_REG_MEASURE, (2600 / 42) & 0x3f);
	if (reg < 0) {
		return reg - 300;
	}

	// set CC source to measure and enable pull-up
	reg = 0; // start with all disconnected
	if (1 == cc) {
		reg |= FUSB302_REG_SWITCHES0_MEAS_CC1 | FUSB302_REG_SWITCHES0_PU_EN1;
	} else if (2 == cc) {
		reg |= FUSB302_REG_SWITCHES0_MEAS_CC2 | FUSB302_REG_SWITCHES0_PU_EN2;
	}
	reg = usb_fusb302_register_write(FUSB302_REG_SWITCHES0, reg);
	if (reg < 0) {
		return reg - 400;
	}

	// get COMP and BC_LEVEL
	sleep_us(100); // before that the result is wrong
	reg = usb_fusb302_register_read(FUSB302_REG_STATUS0);
	if (reg < 0) {
		return reg - 500;
	}

	if (reg & FUSB302_REG_STATUS0_COMP) { // there is no pull down
		// disconnect CC but keep enable pull-up
		reg = 0; // start with all disconnected
		if (1 == cc) {
			reg |= FUSB302_REG_SWITCHES0_PU_EN1;
		} else if (2 == cc) {
			reg |= FUSB302_REG_SWITCHES0_PU_EN2;
		}
		reg = usb_fusb302_register_write(FUSB302_REG_SWITCHES0, reg);
		if (reg < 0) {
			return reg - 600;
		}

		// connect VBUS and compare against 2.6V (any value < 3.3 in MDAC is fine)
		// to find out if a host cable is plugged, we check if the current/voltage we put in CC is feed back to VBUS though Rp
		reg = usb_fusb302_register_write(FUSB302_REG_MEASURE, (2600 / 420) | FUSB302_REG_MEASURE_MEAS_VBUS);
		if (reg < 0) {
			return reg - 700;
		}

		// BC_LEVEL could also be used, but won't allow to differentiate between Ra and Rd by itself
		sleep_us(100); // before that the result is wrong
		reg = usb_fusb302_register_read(FUSB302_REG_STATUS0);
		if (reg < 0) {
			return reg - 800;
		}
		if (reg & FUSB302_REG_STATUS0_COMP) { // there is a voltage (from our pull up)
			r = 3; // indicate Rp
		} else {
			r = 4; // indicate open
		}
	} else { // there is a pull down
		const uint8_t bc_level = reg & FUSB302_REG_STATUS0_BC_LEVEL;
		if (0 == bc_level) { // < 200 mV, less than Ra (~ 330 mV), probably a short
			r = 0; // indicate short
		} else if (1 == bc_level) { // 200 - 660 mV, Ra (300 - 360 mV)
			r = 1; // indicate Ra
		} else { // Rd
			r = 2; // indicate Rd
		}
	}

	// disconnect all CC
	reg = usb_fusb302_disconnect();
	if (reg < 0) {
		return reg;
	}

	return r;
}

int16_t usb_fusb302_pd(uint8_t cc)
{
	if (0 == cc || cc > 2) {
		return -1;
	}
	int16_t reg; // to store register values

	// connected PD CC to MEAS and other CC to VCONN to power eMarker
	// this is not explained in the datasheet, but without connecting to MEAS no data will be received
	reg = usb_fusb302_register_read(FUSB302_REG_SWITCHES0);
	if (reg < 0) {
		return reg - 100;
	}
	reg &= ~(FUSB302_REG_SWITCHES0_MEAS_CC1 | FUSB302_REG_SWITCHES0_MEAS_CC2);
	if (1 == cc) {
		reg |= FUSB302_REG_SWITCHES0_MEAS_CC1;
	} else if (2 == cc) {
		reg |= FUSB302_REG_SWITCHES0_MEAS_CC2;
	}
	reg &= ~(FUSB302_REG_SWITCHES0_VCONN_CC1 | FUSB302_REG_SWITCHES0_VCONN_CC2);
	if (1 == cc) {
		reg |= FUSB302_REG_SWITCHES0_VCONN_CC2;
	} else if (2 == cc) {
		reg |= FUSB302_REG_SWITCHES0_VCONN_CC1;
	}
	reg = usb_fusb302_register_write(FUSB302_REG_SWITCHES0, reg);
	if (reg < 0) {
		return reg - 200;
	}

	// disconnect VBUS and only monitor CC power level
	reg = usb_fusb302_register_read(FUSB302_REG_MEASURE);
	if (reg < 0) {
		return reg - 300;
	}
	reg &= ~FUSB302_REG_MEASURE_MEAS_VBUS;
	reg = usb_fusb302_register_write(FUSB302_REG_MEASURE, reg);
	if (reg < 0) {
		return reg - 400;
	}

	// connect BMC to CC
	reg = usb_fusb302_register_read(FUSB302_REG_SWITCHES1);
	if (reg < 0) {
		return reg - 500;
	}
	reg |= FUSB302_REG_SWITCHES1_AUTO_CRC; // automatically send goodCRC
	reg &= ~(FUSB302_REG_SWITCHES1_TXCC1 | FUSB302_REG_SWITCHES1_TXCC2);
	if (1 == cc) {
		reg |= FUSB302_REG_SWITCHES1_TXCC1;
	} else if (2 == cc) {
		reg |= FUSB302_REG_SWITCHES1_TXCC2;
	}
	reg = usb_fusb302_register_write(FUSB302_REG_SWITCHES1, reg);
	if (reg < 0) {
		return reg - 600;
	}

	// don't ignore SOP'/SOP'' packets (and flush RX)
	reg = FUSB302_REG_CONTROL1_ENSOP2 | FUSB302_REG_CONTROL1_ENSOP1 | FUSB302_REG_CONTROL1_RX_FLUSH;
	reg = usb_fusb302_register_write(FUSB302_REG_CONTROL1, reg);
	if (reg < 0) {
		return reg - 700;
	}

	return 0;
}

int16_t usb_fusb302_send_hard_reset(void)
{
	int16_t reg; // to store register values

	// I don't know why, but only one hard reset can be sent
	// further request to send hard reset don't lead to any transmission
	// this behaviour is not documented in the data sheet
	// resetting the PD logic corrects this state
	reg = usb_fusb302_register_write(FUSB302_REG_RESET, FUSB302_REG_RESET_PD_RESET);
	if (reg < 0) {
		return reg - 100;
	}

	// send hard reset
	reg = usb_fusb302_register_read(FUSB302_REG_CONTROL3);
	if (reg < 0) {
		return reg - 200;
	}
	reg |= FUSB302_REG_CONTROL3_SEND_HARD_RESET;
	reg = usb_fusb302_register_write(FUSB302_REG_CONTROL3, reg);
	if (reg < 0) {
		return reg - 300;
	}

	return 0;
}

int16_t usb_fusb302_rx_flush(void)
{
	int16_t reg; // to store register values

	// flush RX FIFO
	reg = usb_fusb302_register_read(FUSB302_REG_CONTROL1);
	if (reg < 0) {
		return reg - 100;
	}
	reg |= FUSB302_REG_CONTROL1_RX_FLUSH;
	reg = usb_fusb302_register_write(FUSB302_REG_CONTROL1, reg);
	if (reg < 0) {
		return reg - 200;
	}

	return 0;
}

int16_t usb_fusb302_discover_identity_request(void)
{
	int16_t reg; // to store register values
	enum i2c_master_rc i2c_rc; // for I²C return code

	// flush TX FIFO to start new packet
	reg = usb_fusb302_register_read(FUSB302_REG_CONTROL0);
	if (reg < 0) {
		return reg - 100;
	}
	reg |= FUSB302_REG_CONTROL0_TX_FLUSH;
	reg = usb_fusb302_register_write(FUSB302_REG_CONTROL0, reg);
	if (reg < 0) {
		return reg - 200;
	}

	sleep_ms(40); // wait tDiscoverIdentity = 40 ms (PD 3.0 section 6.6.13)

	const uint16_t header = PD_HEADER( // construct header
		0x0f, // type = Vendor_Defined
		0, // power role = sink
		0, // reserved bit for SOP'
		0, // message ID = 0
		1, // number of data objects = 1
		1, // specification revision = 2.0
		0  // extended = false
	);

	const uint32_t object = VDO( // construct VDM header
		0xff00, // SVID = none
		1, // type = structured
		VDO_CMDT(CMDT_INIT) | CMD_DISCOVER_IDENT
	); // request command, discover identity

	const uint8_t packet[] = { // construct packet to sent
		FUSB302_REG_FIFOS_SYNC1, FUSB302_REG_FIFOS_SYNC1, FUSB302_REG_FIFOS_SYNC3, FUSB302_REG_FIFOS_SYNC3, // destination is SOP'
		FUSB302_REG_FIFOS_PACKSYM | (2 + 4), // message length (2 bytes for the header, 4 bytes for the objects)
		(header >> 0) & 0xff, (header >> 8) & 0xff, // header
		(object >> 0) & 0xff, (object >> 8) & 0xff, (object >> 16) & 0xff, (object >> 24) & 0xff, // object
		FUSB302_REG_FIFOS_JAMCRC, // put in CRC
		FUSB302_REG_FIFOS_EOP, // add end of packet
		FUSB302_REG_FIFOS_TXOFF, // indicate to end transmission here
		FUSB302_REG_FIFOS_TXON, // start transmission now (instead of using TXSTART
	};

	const uint8_t addr = FUSB302_REG_FIFOS; // FIFO register address
	i2c_rc = i2c_master_address_write(USB_FUSB302_I2C, USB_FUSB302_ADDR, false, &addr, 1, packet, LENGTH(packet));
	if (I2C_MASTER_RC_NONE != i2c_rc) {
		return -600 - i2c_rc;
	}

	return 0;
}

int16_t usb_fusb302_packet_read(uint8_t* packet)
{
	if (NULL == packet) {
		return -1;
	}
	int16_t reg; // to store register values
	enum i2c_master_rc i2c_rc; // for I²C return code

	// check if some data is available
	reg = usb_fusb302_register_read(FUSB302_REG_STATUS1);
	if (reg < 0) {
		return reg - 100;
	}
	if (reg & FUSB302_REG_STATUS1_RX_EMPTY) {
		return 0;
	}

	// get SOP (start of packet)
	reg = usb_fusb302_register_read(FUSB302_REG_FIFOS);
	if (reg < 0) {
		return reg - 200;
	}
	packet[0] = reg;

	// get message header
	i2c_rc = i2c_master_slave_read(USB_FUSB302_I2C, USB_FUSB302_ADDR, false, &packet[1], 2);
	if (I2C_MASTER_RC_NONE != i2c_rc) {
		return -300 - i2c_rc;
	}
	uint16_t header = (packet[2] << 8) + packet[1];
	if (PD_HEADER_EXT(header)) { // parsing extended messages is not supported currently
		return 3;
	}

	// read rest of message, including CRC
	// I don't know why, but on control messages (message with no object), every subsequent read starts with a token
	i2c_rc = i2c_master_slave_read(USB_FUSB302_I2C, USB_FUSB302_ADDR, false, &packet[3], PD_HEADER_CNT(header) * 4 + 4);
	if (I2C_MASTER_RC_NONE != i2c_rc) {
		return -400 - i2c_rc;
	}

	return 3 + PD_HEADER_CNT(header) * 4 + 4;
}

bool usb_fusb302_packet_checksum(const uint8_t* packet)
{
	if (NULL == packet) {
		return -1;
	}
	uint16_t header = *(uint16_t*)(&packet[1]);
	uint32_t crc = usb_fusb302_crc(&packet[1], PD_HEADER_CNT(header) * 4 + 2);
	uint32_t sum =  *(uint32_t*)(&packet[1 + PD_HEADER_CNT(header) * 4 + 2]);
	return crc == sum;
}

uint32_t usb_fusb302_crc(const uint8_t* data, uint32_t length)
{
	if (NULL == data || 0 == length) {
		return 0;
	}
	uint32_t crc = 0xffffffff;
	crc_reset();

	// calculate for words
	uint32_t cycles = length >> 2; // optimized / 4
	while (cycles--) {
		crc = *(uint32_t *)data;
		data += 4;
		__asm("rbit %0, %1" : "=r" (crc) : "r" (crc)); // reflect input/invert bit order
		crc = crc_calculate(crc);
	}
	__asm("rbit %0, %1" : "=r" (crc) : "r" (crc)); // reflect/invert bit order of output

	// calculate for remaining bytes
	cycles = length & 3; // optimized % 4
	while (cycles--) {
		crc ^= (uint32_t)*data++;
		for (uint8_t b = 0; b < 8; b++) {
			if (crc & 1) {
				crc = (crc >> 1) ^ 0xEDB88320;
			} else {
				crc >>= 1;
			}
		}
	}

	return crc ^ 0xffffffff; // final XOR
}

/* the look-up table method is fast than the hardware assisted calculation for small non-32-bit-aligned data, but uses 1 KB for the LUT
uint32_t usb_fusb302_crc(const uint8_t* data, uint32_t length)
{
	if (NULL == data || 0 == length) {
		return 0;
	}
	// CRC-32 Look-Up Table: input reflected, result reflected, polynomial 0x4C11DB7, initial value: 0xFFFFFFFF, final xor value: 0xFFFFFFFF, table is reflected, generated by [pycrc](https://pycrc.org/)
	static const uint32_t crc_table[256] = {
		0x00000000, 0x77073096, 0xee0e612c, 0x990951ba, 0x076dc419, 0x706af48f, 0xe963a535, 0x9e6495a3,
		0x0edb8832, 0x79dcb8a4, 0xe0d5e91e, 0x97d2d988, 0x09b64c2b, 0x7eb17cbd, 0xe7b82d07, 0x90bf1d91,
		0x1db71064, 0x6ab020f2, 0xf3b97148, 0x84be41de, 0x1adad47d, 0x6ddde4eb, 0xf4d4b551, 0x83d385c7,
		0x136c9856, 0x646ba8c0, 0xfd62f97a, 0x8a65c9ec, 0x14015c4f, 0x63066cd9, 0xfa0f3d63, 0x8d080df5,
		0x3b6e20c8, 0x4c69105e, 0xd56041e4, 0xa2677172, 0x3c03e4d1, 0x4b04d447, 0xd20d85fd, 0xa50ab56b,
		0x35b5a8fa, 0x42b2986c, 0xdbbbc9d6, 0xacbcf940, 0x32d86ce3, 0x45df5c75, 0xdcd60dcf, 0xabd13d59,
		0x26d930ac, 0x51de003a, 0xc8d75180, 0xbfd06116, 0x21b4f4b5, 0x56b3c423, 0xcfba9599, 0xb8bda50f,
		0x2802b89e, 0x5f058808, 0xc60cd9b2, 0xb10be924, 0x2f6f7c87, 0x58684c11, 0xc1611dab, 0xb6662d3d,
		0x76dc4190, 0x01db7106, 0x98d220bc, 0xefd5102a, 0x71b18589, 0x06b6b51f, 0x9fbfe4a5, 0xe8b8d433,
		0x7807c9a2, 0x0f00f934, 0x9609a88e, 0xe10e9818, 0x7f6a0dbb, 0x086d3d2d, 0x91646c97, 0xe6635c01,
		0x6b6b51f4, 0x1c6c6162, 0x856530d8, 0xf262004e, 0x6c0695ed, 0x1b01a57b, 0x8208f4c1, 0xf50fc457,
		0x65b0d9c6, 0x12b7e950, 0x8bbeb8ea, 0xfcb9887c, 0x62dd1ddf, 0x15da2d49, 0x8cd37cf3, 0xfbd44c65,
		0x4db26158, 0x3ab551ce, 0xa3bc0074, 0xd4bb30e2, 0x4adfa541, 0x3dd895d7, 0xa4d1c46d, 0xd3d6f4fb,
		0x4369e96a, 0x346ed9fc, 0xad678846, 0xda60b8d0, 0x44042d73, 0x33031de5, 0xaa0a4c5f, 0xdd0d7cc9,
		0x5005713c, 0x270241aa, 0xbe0b1010, 0xc90c2086, 0x5768b525, 0x206f85b3, 0xb966d409, 0xce61e49f,
		0x5edef90e, 0x29d9c998, 0xb0d09822, 0xc7d7a8b4, 0x59b33d17, 0x2eb40d81, 0xb7bd5c3b, 0xc0ba6cad,
		0xedb88320, 0x9abfb3b6, 0x03b6e20c, 0x74b1d29a, 0xead54739, 0x9dd277af, 0x04db2615, 0x73dc1683,
		0xe3630b12, 0x94643b84, 0x0d6d6a3e, 0x7a6a5aa8, 0xe40ecf0b, 0x9309ff9d, 0x0a00ae27, 0x7d079eb1,
		0xf00f9344, 0x8708a3d2, 0x1e01f268, 0x6906c2fe, 0xf762575d, 0x806567cb, 0x196c3671, 0x6e6b06e7,
		0xfed41b76, 0x89d32be0, 0x10da7a5a, 0x67dd4acc, 0xf9b9df6f, 0x8ebeeff9, 0x17b7be43, 0x60b08ed5,
		0xd6d6a3e8, 0xa1d1937e, 0x38d8c2c4, 0x4fdff252, 0xd1bb67f1, 0xa6bc5767, 0x3fb506dd, 0x48b2364b,
		0xd80d2bda, 0xaf0a1b4c, 0x36034af6, 0x41047a60, 0xdf60efc3, 0xa867df55, 0x316e8eef, 0x4669be79,
		0xcb61b38c, 0xbc66831a, 0x256fd2a0, 0x5268e236, 0xcc0c7795, 0xbb0b4703, 0x220216b9, 0x5505262f,
		0xc5ba3bbe, 0xb2bd0b28, 0x2bb45a92, 0x5cb36a04, 0xc2d7ffa7, 0xb5d0cf31, 0x2cd99e8b, 0x5bdeae1d,
		0x9b64c2b0, 0xec63f226, 0x756aa39c, 0x026d930a, 0x9c0906a9, 0xeb0e363f, 0x72076785, 0x05005713,
		0x95bf4a82, 0xe2b87a14, 0x7bb12bae, 0x0cb61b38, 0x92d28e9b, 0xe5d5be0d, 0x7cdcefb7, 0x0bdbdf21,
		0x86d3d2d4, 0xf1d4e242, 0x68ddb3f8, 0x1fda836e, 0x81be16cd, 0xf6b9265b, 0x6fb077e1, 0x18b74777,
		0x88085ae6, 0xff0f6a70, 0x66063bca, 0x11010b5c, 0x8f659eff, 0xf862ae69, 0x616bffd3, 0x166ccf45,
		0xa00ae278, 0xd70dd2ee, 0x4e048354, 0x3903b3c2, 0xa7672661, 0xd06016f7, 0x4969474d, 0x3e6e77db,
		0xaed16a4a, 0xd9d65adc, 0x40df0b66, 0x37d83bf0, 0xa9bcae53, 0xdebb9ec5, 0x47b2cf7f, 0x30b5ffe9,
		0xbdbdf21c, 0xcabac28a, 0x53b39330, 0x24b4a3a6, 0xbad03605, 0xcdd70693, 0x54de5729, 0x23d967bf,
		0xb3667a2e, 0xc4614ab8, 0x5d681b02, 0x2a6f2b94, 0xb40bbe37, 0xc30c8ea1, 0x5a05df1b, 0x2d02ef8d
	};
	uint32_t crc = 0xffffffff;
	const unsigned char *d = (const unsigned char *)data;
	unsigned int tbl_idx;

	while (length--) {
		tbl_idx = (crc ^ *d) & 0xff;
		crc = (crc_table[tbl_idx] ^ (crc >> 8)) & 0xffffffff;
		d++;
	}
	return (crc & 0xffffffff) ^ 0xffffffff;
}
*/
