/* This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/** library to communicate using USB-C controller FUSB302
 *  @file
 *  @author King Kévin <kingkevin@cuvoodoo.info>
 *  @date 2020
 *  @note peripherals used: I²C @ref usb_fusb302_i2c
 *  @note the interrupt pin is not handled by this library
 */
#pragma once

/** setup I²C bus to communicate with FUSB302
 *  @return 0 when communication to FUSB302 is successful, < 0 else
 */
int16_t usb_fusb302_setup(void);
/** disconnect all from CC and VBUS lines
 *  @return 0 on success, < 0 on error
 */
int16_t usb_fusb302_disconnect(void);
/** read register from FUSB302
 *  @param[in] address register address
 *  @return register value on success, < 0 on error
 */
int16_t usb_fusb302_register_read(uint8_t address);
/** write register to FUSB302
 *  @param[in] address register address
 *  @param[in] value register value to write
 *  @return 0 on success, < 0 on error
 */
int16_t usb_fusb302_register_write(uint8_t address, uint8_t value);
/** get version of FUSB302
 *  @return version A-Z, ! if I²C error, ? if undefined
 */
char usb_fusb302_version(void);
/** get revision of FUSB302
 *  @return version A-Z, ! if I²C error
 */
char usb_fusb302_revision(void);
/** get VBUS voltage
 *  @param[in] source 0: VBUS, 1: CC1, 2: CC2
 *  @return VBUS voltage, in mV, or < 0 on error
 *  @note this is not the exact voltage. the actual voltage can be up to 420 mV for VBUS or 42 mV for CC above the returned value for values < 26460 or 2646 mV, anything above else
 */
int32_t usb_fusb302_voltage(uint8_t source);
/** find out which resistor is connected to a CC line
 *  @param[in] cc CC channel to probe (1 or 2)
 *  @return 0 for short to ground, 1 for Ra, 2 for Rd, 3 for Rp, 4 for open, < 0 on error
 *  @warning disconnects all from CC/VBUS at the end
 */
int16_t usb_fusb302_r(uint8_t cc);
/** connect BMC to CC line for Power Delivery communication
 *  @param[in] cc CC line to connected to BMC
 *  @return 0 on success, < 0 on error
 *  @note connects VCONN on other CC line
 *  @warning flushes RX FIFO
 */
int16_t usb_fusb302_pd(uint8_t cc);
/** send hard reset
 *  @return 0 on success, < 0 on error
 *  @warning PD channel must be configured
 */
int16_t usb_fusb302_send_hard_reset(void);
/** fluch RX FIFO
 *  @return 0 on success, < 0 on error
 *  @warning PD channel must be configured
 */
int16_t usb_fusb302_rx_flush(void);
/** send discover identity request (to SOP')
 *  @return 0 on success, < 0 on error
 *  @warning PD channel must be configured
 */
int16_t usb_fusb302_discover_identity_request(void);
/** read received packet
 *  @param[out] packet buffer for packet data (up to 1 + 2 +  7 * 4 + 4 = 35 bytes long)
 *  @return number of bytes read, < 0 on error
 *  @warning PD channel must be configured
 */
int16_t usb_fusb302_packet_read(uint8_t* packet);
/** calculate CRC for paylod
 *  @param[in] data payload data for which to calculate the checksum
 *  @param[in] length length of data
 *  @return CRC checksum
 *  @implements CRC-32-IEEE also used for PD 3.0 specification (most common CRC-32)
 */
uint32_t usb_fusb302_crc(const uint8_t* data, uint32_t length);
/** verify CRC checksum of packet
 *  @param[in] packet packet for which to verify the checksum (CRC is included)
 *  @return if the calculated checksum matches the one included in the packet
 */
bool usb_fusb302_packet_checksum(const uint8_t* packet);

/** FUSB302 register addresses and bit offset */
#define FUSB302_REG_DEVICE_ID 0x01
#define FUSB302_REG_SWITCHES0 0x02
#define FUSB302_REG_SWITCHES0_PDWN1 (1 << 0)
#define FUSB302_REG_SWITCHES0_PDWN2 (1 << 1)
#define FUSB302_REG_SWITCHES0_MEAS_CC1 (1 << 2)
#define FUSB302_REG_SWITCHES0_MEAS_CC2 (1 << 3)
#define FUSB302_REG_SWITCHES0_VCONN_CC1 (1 << 4)
#define FUSB302_REG_SWITCHES0_VCONN_CC2 (1 << 5)
#define FUSB302_REG_SWITCHES0_PU_EN1 (1 << 6)
#define FUSB302_REG_SWITCHES0_PU_EN2 (1 << 7)
#define FUSB302_REG_SWITCHES1 0x03
#define FUSB302_REG_SWITCHES1_TXCC1 (1 << 0)
#define FUSB302_REG_SWITCHES1_TXCC2 (1 << 1)
#define FUSB302_REG_SWITCHES1_AUTO_CRC (1 << 2)
#define FUSB302_REG_SWITCHES1_DATAROLE (1 << 4)
#define FUSB302_REG_SWITCHES1_POWERROLE (1 << 7)
#define FUSB302_REG_MEASURE 0x04
#define FUSB302_REG_MEASURE_MEAS_VBUS (1 << 6)
#define FUSB302_REG_SLICE 0x05
#define FUSB302_REG_CONTROL0 0x06
#define FUSB302_REG_CONTROL0_TX_START (1 << 0)
#define FUSB302_REG_CONTROL0_AUTO_PRE (1 << 1)
#define FUSB302_REG_CONTROL0_HOST_CUR_MASK (0x3 << 2)
#define FUSB302_REG_CONTROL0_HOST_CUR_0UA (0x3 << 2)
#define FUSB302_REG_CONTROL0_HOST_CUR_80UA (0x1 << 2)
#define FUSB302_REG_CONTROL0_HOST_CUR_180UA (0x2 << 2)
#define FUSB302_REG_CONTROL0_HOST_CUR_330UA (0x3 << 2)
#define FUSB302_REG_CONTROL0_INT_MASK (1 << 5)
#define FUSB302_REG_CONTROL0_TX_FLUSH (1 << 6)
#define FUSB302_REG_CONTROL1 0x07
#define FUSB302_REG_CONTROL1_ENSOP1 (1 << 0)
#define FUSB302_REG_CONTROL1_ENSOP2 (1 << 1)
#define FUSB302_REG_CONTROL1_RX_FLUSH (1 << 2)
#define FUSB302_REG_CONTROL1_BIST_MODE2 (1 << 4)
#define FUSB302_REG_CONTROL1_ENSOP1DB (1 << 5)
#define FUSB302_REG_CONTROL1_ENSOP2DB (1 << 6)
#define FUSB302_REG_CONTROL2 0x08
#define FUSB302_REG_CONTROL2_TOGGLE (1 << 0)
#define FUSB302_REG_CONTROL2_WAKE_EN (1 << 3)
#define FUSB302_REG_CONTROL2_TOG_RD_ONLY (1 << 5)
#define FUSB302_REG_CONTROL3 0x09
#define FUSB302_REG_CONTROL3_AUTO_RETRY (1 << 0)
#define FUSB302_REG_CONTROL3_AUTO_SOFTRESET (1 << 3)
#define FUSB302_REG_CONTROL3_AUTO_HARDRESET (1 << 4)
#define FUSB302_REG_CONTROL3_SEND_HARD_RESET (1 << 6)
#define FUSB302_REG_MASK 0x0a
#define FUSB302_REG_MASK_M_BC_LVL (1 << 0)
#define FUSB302_REG_MASK_M_COLLISION (1 << 1)
#define FUSB302_REG_MASK_M_WAKE (1 << 2)
#define FUSB302_REG_MASK_M_ALERT (1 << 3)
#define FUSB302_REG_MASK_M_CRC_CHK (1 << 4)
#define FUSB302_REG_MASK_M_COMP_CHNG (1 << 5)
#define FUSB302_REG_MASK_M_ACTIVITY (1 << 6)
#define FUSB302_REG_MASK_M_VBUSOK (1 << 7)
#define FUSB302_REG_POWER 0x0b
#define FUSB302_REG_POWER_BANDGAP_WAKEUP (1 << 0)
#define FUSB302_REG_POWER_RECEIVER_REFERENCES (1 << 1)
#define FUSB302_REG_POWER_MEASURE (1 << 2)
#define FUSB302_REG_POWER_OSCILLATOR (1 << 3)
#define FUSB302_REG_RESET 0x0c
#define FUSB302_REG_RESET_SW_RES (1 << 0)
#define FUSB302_REG_RESET_PD_RESET (1 << 1)
#define FUSB302_REG_OCPREG 0x0d
#define FUSB302_REG_MASKA 0x0e
#define FUSB302_REG_MASKA_M_HARDRST (1 << 0)
#define FUSB302_REG_MASKA_M_SOFTRST (1 << 1)
#define FUSB302_REG_MASKA_M_TXSENT (1 << 2)
#define FUSB302_REG_MASKA_M_HARDSENT (1 << 3)
#define FUSB302_REG_MASKA_M_RETRYFAIL (1 << 4)
#define FUSB302_REG_MASKA_M_SOFTFAIL (1 << 5)
#define FUSB302_REG_MASKA_M_TOGDONE (1 << 6)
#define FUSB302_REG_MASKA_M_OCP_TEMP (1 << 7)
#define FUSB302_REG_MASKB 0x0f
#define FUSB302_REG_MASKB_M_GCRCSENT (1 << 0)
#define FUSB302_REG_CONTROL4 0x10
#define FUSB302_REG_CONTROL4_TOG_USRC_EXIT (1 << 0)
#define FUSB302_REG_STATUS0A 0x3c
#define FUSB302_REG_STATUS0A_HARDRST (1 << 0)
#define FUSB302_REG_STATUS0A_SOFTRST (1 << 1)
#define FUSB302_REG_STATUS0A_RETRYFAIL (1 << 4)
#define FUSB302_REG_STATUS0A_SOFTFAIL (1 << 5)
#define FUSB302_REG_STATUS1A 0x3d
#define FUSB302_REG_STATUS1A_RXSOP (1 << 0)
#define FUSB302_REG_STATUS1A_RXSOP1DB (1 << 1)
#define FUSB302_REG_STATUS1A_RXSOP2DB (1 << 2)
#define FUSB302_REG_INTERRUPTA 0x3e
#define FUSB302_REG_INTERRUPTA_I_HARDRST (1 << 0)
#define FUSB302_REG_INTERRUPTA_I_SOFTRST (1 << 1)
#define FUSB302_REG_INTERRUPTA_I_TXSENT (1 << 2)
#define FUSB302_REG_INTERRUPTA_I_HARDSENT (1 << 3)
#define FUSB302_REG_INTERRUPTA_I_RETRYFAIL (1 << 4)
#define FUSB302_REG_INTERRUPTA_I_SOFTFAIL (1 << 5)
#define FUSB302_REG_INTERRUPTA_I_TOGDONE (1 << 6)
#define FUSB302_REG_INTERRUPTA_I_OCP_TEMP (1 << 7)
#define FUSB302_REG_INTERRUPTB 0x3f
#define FUSB302_REG_INTERRUPTB_I_GCRCSENT (1 << 0)
#define FUSB302_REG_STATUS0 0x40
#define FUSB302_REG_STATUS0_BC_LEVEL 0x3
#define FUSB302_REG_STATUS0_WAKE (1 << 2)
#define FUSB302_REG_STATUS0_ALERT (1 << 3)
#define FUSB302_REG_STATUS0_CRC_CHK (1 << 4)
#define FUSB302_REG_STATUS0_COMP (1 << 5)
#define FUSB302_REG_STATUS0_ACTIVITY (1 << 6)
#define FUSB302_REG_STATUS0_VBUSOK (1 << 7)
#define FUSB302_REG_STATUS1 0x41
#define FUSB302_REG_STATUS1_OCP (1 << 0)
#define FUSB302_REG_STATUS1_OVRTEMP (1 << 1)
#define FUSB302_REG_STATUS1_TX_FULL (1 << 2)
#define FUSB302_REG_STATUS1_TX_EMPTY (1 << 3)
#define FUSB302_REG_STATUS1_RX_FULL (1 << 4)
#define FUSB302_REG_STATUS1_RX_EMPTY (1 << 5)
#define FUSB302_REG_STATUS1_RXSOP1 (1 << 6)
#define FUSB302_REG_STATUS1_RXSOP2 (1 << 7)
#define FUSB302_REG_INTERRUPT 0x42
#define FUSB302_REG_INTERRUPT_I_BC_LVL (1 << 0)
#define FUSB302_REG_INTERRUPT_I_COLLISION (1 << 1)
#define FUSB302_REG_INTERRUPT_I_WAKE (1 << 2)
#define FUSB302_REG_INTERRUPT_I_ALERT (1 << 3)
#define FUSB302_REG_INTERRUPT_I_CRC_CHK (1 << 4)
#define FUSB302_REG_INTERRUPT_I_COMP_CHNG (1 << 5)
#define FUSB302_REG_INTERRUPT_I_ACTIVITY (1 << 6)
#define FUSB302_REG_INTERRUPt_I_VBUSOK (1 << 7)
#define FUSB302_REG_FIFOS 0x43
#define FUSB302_REG_FIFOS_TXON 0xA1
#define FUSB302_REG_FIFOS_SYNC1 0x12
#define FUSB302_REG_FIFOS_SYNC2 0x13
#define FUSB302_REG_FIFOS_SYNC3 0x1B
#define FUSB302_REG_FIFOS_RESET1 0x15
#define FUSB302_REG_FIFOS_RESET2 0x16
#define FUSB302_REG_FIFOS_PACKSYM 0x80
#define FUSB302_REG_FIFOS_JAMCRC 0xFF
#define FUSB302_REG_FIFOS_EOP 0x14
#define FUSB302_REG_FIFOS_TXOFF 0xFE
